# -*- coding: utf-8 -*-
# name: twopiece.shape.py
# author: D.Santiago
# https://www.linkedin.com/in/dialidsantiago/
# @Quant_Girl
# --
# coding: utf-8


import scipy.stats

from twopiece.double import tpd_continuous
from twopiece.sinharcsinh import ssas
from twopiece.utils import display_dist


class TwoPieceShape(tpd_continuous):

    def __init__(self, f, loc=0.0, sigma=1.0, shape1=None, shape2=None):
        tpd_continuous.__init__(self, f, loc, sigma, sigma, None, None, shape1, shape2, 'boe')


class tpshastudent(TwoPieceShape):

    def __init__(self, loc=0.0, sigma=1.0, shape1=3.0, shape2=3.0):
        TwoPieceShape.__init__(self, scipy.stats.t, loc, sigma, shape1, shape2)


class tpshagennorm(TwoPieceShape):

    def __init__(self, loc=0.0, sigma=1.0, shape1=3.0, shape2=3.0):
        TwoPieceShape.__init__(self, scipy.stats.gennorm, loc, sigma, shape1, shape2)


class tpshasas(TwoPieceShape):

    def __init__(self, loc=0.0, sigma=None, shape1=3.0, shape2=3.0):
        TwoPieceShape.__init__(self, ssas, loc, sigma, shape1, shape2)


def display_tpshape(tpd='All', loc=0.0, sigma=1.0, shape1=2.0, shape2=6.0, show='random_sample', xlim=None):
    if tpd in ['All', 'student']:
        z = tpshastudent(loc=loc, sigma=sigma, shape1=shape1, shape2=shape2)
        display_dist(dist=z, color='dodgerblue', bound=True, name='dtpshstudent', show=show, xlim=xlim)

    if tpd in ['All', 'gennorm']:
        z = tpshagennorm(loc=loc, sigma=sigma, shape1=shape1, shape2=shape2)
        display_dist(dist=z, color='gold', name='tpshagennorm', show=show, xlim=xlim)

    if tpd in ['All', 'Sinhasinh']:
        z = tpshasas(loc=loc, sigma=sigma, shape1=shape1, shape2=shape2)
        display_dist(dist=z, color='deeppink', name='dtshapsas', show=show, xlim=xlim)

    return None




# -*- coding: utf-8 -*-
# name: twopiece.double.py
# author: D.Santiago
# https://www.linkedin.com/in/dialidsantiago/
# @Quant_Girl
# --
# coding: utf-8


import scipy.stats
from numpy import isscalar, asarray, random, sum, empty

from twopiece.sinharcsinh import ssas
from twopiece.utils import get_sigma1_sigma2, display_dist


def pdf_tpd_generic(x, pdf1, pdf2, loc, sigma1, sigma2, epsilon):
    """
    Probability density function at x of the defined two piece distribution.

    :param pdf1:
    :param pdf2:
    :param x: array like
    :param pdf: a probability density function from a symmetric distribution defined on R.
    :param loc: location parameter
    :param sigma1: scale parameter
    :param sigma2: scale parameter
    :param epsilon: shape parameter
    :return: pdf of the defined two piece in x

    """

    if sigma1 * sigma2 <= 0:
        raise AssertionError('Scale parameters must be positive.')

    aux1 = 2 * epsilon / sigma1
    aux2 = 2 * (1 - epsilon) / sigma2

    if isscalar(x):

        if x < loc:

            output = aux1 * pdf1((x - loc) / sigma1)

        else:

            output = aux2 * pdf2((x - loc) / sigma2)
    else:

        x = asarray(x)
        output = empty(x.size)
        index = x < loc
        output[index] = aux1 * pdf1((x[index] - loc) / sigma1)
        index = x >= loc
        output[index] = aux2 * pdf2((x[index] - loc) / sigma2)

    return output


def cdf_tpd_generic(x, cdf1, cdf2, loc, sigma1, sigma2, epsilon):
    """
    Cumulative Density Function at x of the defined two piece distribution.

    :param x: array like
    :param cdf: a cumulative density function from a symmetric distribution defined on R.
    :param loc: location parameter
    :param sigma1: scale parameter
    :param sigma2: scale parameter
    :param epsilon: shape parameter
    :return:

    """

    if sigma1 * sigma2 <= 0:
        raise AssertionError('Scale parameters must be positive.')

    if isscalar(x):

        if x < loc:
            output = 2 * epsilon * cdf1((x - loc) / sigma1)
        else:
            output = epsilon + (1 - epsilon) * (2 * cdf2((x - loc) / sigma2) - 1)

    else:
        x = asarray(x)
        output = empty(x.size)
        index = x < loc
        output[index] = 2 * epsilon * cdf1((x[index] - loc) / sigma1)
        index = x >= loc
        output[index] = epsilon + (1 - epsilon) * (2 * cdf2((x[index] - loc) / sigma2) - 1)

    return output


def qqf_tpd_generic(q, qqf1, qqf2, loc, sigma1, sigma2, epsilon):
    """
    Quantile Function at q of the defined two piece distribution.

    :param q: array like
    :param qqf: a quantile function (ppf) from a symmetric distribution defined on R.
    :param loc: location parameter
    :param sigma1: scale parameter
    :param sigma2: scale parameter
    :param epsilon: shape parameter
    :return:
    """

    if sigma1 * sigma2 <= 0:
        raise AssertionError('Scale parameters must be positive.')

    if isscalar(q):

        if q > 1 or q < 0:
            raise AssertionError('Quantile Function is defined on (0,1).')

        if q <= epsilon:

            output = loc + sigma1 * qqf1((0.5 / epsilon) * q)

        else:
            aux = 0.5 * ((q - epsilon) / (1 - epsilon) + 1)
            output = loc + sigma2 * qqf2(aux)

    else:

        q = asarray(q)

        if sum((q > 1) | (q < 0)) > 0:
            raise AssertionError('Quantile Function is defined on (0,1).')

        output = empty(q.size)
        index = q <= epsilon
        output[index] = loc + sigma1 * qqf1((0.5 / epsilon) * q[index])
        index = q > epsilon
        aux = 0.5 * ((q[index] - epsilon) / (1 - epsilon) + 1)
        output[index] = loc + sigma2 * qqf2(aux)

    return output


def random_tpd_sample(size, qqf1, qqf2, loc, sigma1, sigma2, epsilon):
    """
    Random Sample Generation

    :param size: integer, sample size
    :param qqf: a quantile function (ppf) from a symmetric distribution defined on R
    :param loc: location parameter
    :param sigma1: scale parameter
    :param sigma2: scale parameter
    :param epsilon: shape parameter
    :return:
    """
    if not isinstance(size, int):
        raise TypeError('Sample size must be integer.')

    if sigma1 * sigma2 <= 0:
        raise AssertionError('Scale parameters must be positive.')

    alpha = random.rand(size)

    if isscalar(alpha):

        if alpha <= epsilon:

            qq = loc + sigma1 * qqf1(0.5 * alpha / epsilon)

        else:
            qq = loc + sigma2 * qqf2(0.5 * ((alpha - epsilon) / (1 - epsilon) + 1))

    else:

        alpha = asarray(alpha)
        qq = empty(alpha.size)
        index = alpha <= epsilon
        qq[index] = loc + sigma1 * qqf1(0.5 * alpha[index] * (1 / epsilon))
        index = alpha > epsilon
        qq[index] = loc + sigma2 * qqf2(0.5 * ((alpha[index] - epsilon) / (1 - epsilon) + 1))

    return qq


class TwoPieceDouble:

    def __init__(self, f, loc, sigma1, sigma2, sigma, gamma, shape1, shape2, kind):

        """

        :param f: continuous symmetric distribution with support on R
        :param loc: location parameter
        :param sigma1: scale parameter
        :param sigma2: scale parameter
        :param sigma: scale parameter
        :param gamma: skewness or asymmetry parameter
        :param shape1: shape parameter
        :param shape2: shape parameter
        :param kind: Parametrisation
        """

        if all(v is None for v in {sigma1, sigma2, sigma, gamma}):
            raise AssertionError('Expected either (sigma1, sigma2) or (sigma, gamma).')

        if kind not in {'inverse_scale', 'epsilon_skew', 'percentile', 'boe'}:
            raise AssertionError('Invalid Parametrisation')

        self.f = f
        self.loc = loc
        self.sigma = sigma
        self.gamma = gamma
        self.shape1 = shape1
        self.shape2 = shape2
        self.kind = kind

        if sigma1 and sigma2:

            self.sigma1 = sigma1
            self.sigma2 = sigma2

        else:

            try:

                sigma1, sigma2 = get_sigma1_sigma2(self.sigma, self.gamma, self.kind)
                self.sigma1 = sigma1
                self.sigma2 = sigma2

            except BaseException:
                print('Missing or Invalid Arguments')


class tpd_continuous(TwoPieceDouble):

    def __init__(self, f, loc, sigma1, sigma2, sigma, gamma, shape1, shape2, kind):
        super().__init__(f, loc, sigma1, sigma2, sigma, gamma, shape1, shape2, kind)

        self.f1 = self.f(self.shape1)
        self.f2 = self.f(self.shape2)
        self.epsilon = self.sigma1 * self.f2.pdf(0) / (self.sigma1 * self.f2.pdf(0) + self.sigma2 * self.f1.pdf(0))

    def pdf(self, x):
        s = pdf_tpd_generic(x, self.f1.pdf, self.f2.pdf, self.loc, self.sigma1, self.sigma2, self.epsilon)

        return s

    def cdf(self, x):
        s = cdf_tpd_generic(x, self.f1.cdf, self.f2.cdf, self.loc, self.sigma1, self.sigma2, self.epsilon)

        return s

    def ppf(self, x):
        qq = qqf_tpd_generic(x, self.f1.ppf, self.f2.ppf, self.loc, self.sigma1, self.sigma2, self.epsilon)

        return qq

    def random_sample(self, size):
        sample = random_tpd_sample(size, self.f1.ppf, self.f2.ppf, self.loc, self.sigma1, self.sigma2, self.epsilon)

        return sample

class dtpstudent(tpd_continuous):

    def __init__(self, loc=0.0, sigma1=None, sigma2=None, sigma=None, gamma=None, shape1=None, shape2=None, kind='boe'):
        tpd_continuous.__init__(self, scipy.stats.t, loc, sigma1, sigma2, sigma, gamma, shape1, shape2, kind)


class dtpgennorm(tpd_continuous):

    def __init__(self, loc=0.0, sigma1=None, sigma2=None, sigma=None, gamma=None, shape1=None, shape2=None, kind='boe'):
        tpd_continuous.__init__(self, scipy.stats.gennorm, loc, sigma1, sigma2, sigma, gamma, shape1, shape2, kind)


class dtpsas(tpd_continuous):

    def __init__(self, loc=0.0, sigma1=None, sigma2=None, sigma=None, gamma=None, shape1=None, shape2=None, kind='boe'):
        tpd_continuous.__init__(self, ssas, loc, sigma1, sigma2, sigma, gamma, shape1, shape2, kind)


def display_dtp(tpd='All', loc=0.0, sigma1=1.0, sigma2=0.5, shape1=2.0, shape2=6.0, show='random_sample', xlim=None):
    if tpd in ['All', 'dtpstudent']:
        z = dtpstudent(loc=loc, sigma1=sigma1, sigma2=sigma2, shape1=shape1, shape2=shape2)
        display_dist(dist=z, color='dodgerblue', bound=True, name='dtpstudent', show=show, xlim=xlim)

    if tpd in ['All', 'dtpgennorm']:
        z = dtpgennorm(loc=loc, sigma1=sigma1, sigma2=sigma2, shape1=shape1, shape2=shape2)
        display_dist(dist=z, color='gold', name='dtpgennorm', show=show, xlim=xlim)

    if tpd in ['All', 'dtpsas']:
        z = dtpsas(loc=loc, sigma1=sigma1, sigma2=sigma2, shape1=shape1, shape2=shape2)
        display_dist(dist=z, color='deeppink', name='dtpsas', show=show, xlim=xlim)

    return None



import unittest
from twopiece.scale import tpnorm, get_sigma1_sigma2
import numpy as np
from parameterized import parameterized

class TestTwoPiece(unittest.TestCase):

    @parameterized.expand([
                          [0.0, 1.0, 0.0],
                          [0.0, 1.0, 1.0], [0.0, 1.0, -1.0],
                          [0.0, 1.0, 2.0], [0.0, 1.0, -2.0],])
    def test_boe_parametrisation(self, mu, sigma, gamma):

        sigma1, sigma2 = get_sigma1_sigma2(sigma, gamma, kind='boe')
        dist = tpnorm(mu, sigma1=sigma1, sigma2=sigma2)
        dist_boe = tpnorm(loc=mu, sigma=sigma, gamma=gamma, kind='boe')

        x = np.arange(-12, 12, 0.1)
        pdf = dist.pdf(x)
        pdf_boe = dist_boe.pdf(x)

        for y, y_boe in zip(pdf, pdf_boe):
            self.assertEqual(y, y_boe)  # add assertion here

        mean_theoretical = mu + np.sqrt(2/np.pi)*(sigma2-sigma1)
        gamma_theoretical = mean_theoretical-mu
        self.assertAlmostEqual(gamma, gamma_theoretical)


if __name__ == '__main__':
    unittest.main()



from math import asinh, cosh, sqrt, sinh

import scipy.stats
from numpy import random, vectorize


def _pdf_instance(x, pdf, loc, scale, delta, epsilon):
    z = (x - loc) / scale
    output = (delta / scale) * pdf(sinh(delta * asinh(z) - epsilon)) * cosh(delta * asinh(z) - epsilon) / sqrt(
        1 + z * z)
    return output


def _cdf_instance(x, cdf, loc, scale, delta, epsilon):
    z = (x - loc) / scale
    output = cdf(sinh(delta * asinh(z) - epsilon))

    return output


def _qqf_instance(q, qqf, loc, scale, delta, epsilon):
    output = loc + scale * sinh((asinh(qqf(q)) + epsilon) / delta)
    return output


def _random_sample(size, qqf, loc, scale, delta, epsilon):
    alpha = random.rand(size)
    qqfv = vectorize(_qqf_instance)

    return qqfv(alpha, qqf, loc, scale, delta, epsilon)


class SinhArcsinh:

    def __init__(self, f, loc, scale, delta, epsilon):
        self.f = f
        self.loc = loc
        self.scale = scale
        self.delta = delta
        self.epsilon = epsilon

    def pdf(self, x):
        _pdf_vector = vectorize(_pdf_instance)
        s = _pdf_vector(x, self.f.pdf, self.loc, self.scale, self.delta, self.epsilon)
        return s

    def cdf(self, x):
        _cdf_vector = vectorize(_cdf_instance)
        s = _cdf_vector(x, self.f.cdf, self.loc, self.scale, self.delta, self.epsilon)
        return s

    def ppf(self, q):
        _qqf_vector = vectorize(_qqf_instance)
        x = _qqf_vector(q, self.f.ppf, self.loc, self.scale, self.delta, self.epsilon)

        return x

    def random_sample(self, size):
        sample = _random_sample(size, self.f.ppf, self.loc, self.scale, self.delta, self.epsilon)
        return sample


class sinhasinh(SinhArcsinh):

    def __init__(self, loc=0.0, scale=1.0, delta=1.0, epsilon=0.0):
        SinhArcsinh.__init__(self, scipy.stats.norm, loc, scale, delta, epsilon)


class ssas(SinhArcsinh):

    def __init__(self, delta=1.0):
        SinhArcsinh.__init__(self, f=scipy.stats.norm, loc=0.0, scale=1.0, delta=delta, epsilon=0.0)



import math

import matplotlib.pyplot as plt
from numpy import min, max, arange
from seaborn import distplot
from seaborn import set

set(style='whitegrid', rc={"grid.linewidth": 0.75, "figure.figsize": (9, 6)})


def get_sigma1_sigma2(sigma, gamma, kind):
    """
    Gets the scale parameters sigma1, sigma2 from sigma and gamma

    :param sigma: scale parameter
    :param gamma: skewness or asymmetry parameter
    :param kind: Parametrisation name
    :return:
    """

    if kind == 'inverse_scale':

        if gamma < 0:
            raise AssertionError("Gamma parameter must be positive")
        sigma1 = sigma / gamma
        sigma2 = sigma * gamma

    elif kind == 'epsilon_skew':

        if gamma >= 1 or gamma <= -1:
            raise AssertionError("Gamma parameter must be positive")

        sigma1 = sigma * (1 + gamma)
        sigma2 = sigma * (1 - gamma)

    elif kind == 'percentile':

        if gamma >= 1 or gamma <= 0:
            raise AssertionError("Gamma parameter must be in (0,1)")

        sigma1 = sigma * gamma
        sigma2 = sigma * (1 - gamma)

    else:
        if gamma >= 1 or gamma <= -1:
            raise AssertionError("Skewness parameters must be in (-1, 1).")

        sigma1 = sigma / math.sqrt(1 + gamma)
        sigma2 = sigma / math.sqrt(1 - gamma)

    return sigma1, sigma2


def display_dist(dist, name='', color='dodgerblue', bound=False, show='random_sample', xlim=None):
    """
    Shows graphs for a given two piece distribution
    :param dist: distribution instance
    :param name: string, name
    :param color: string, color
    :param bound: boolean, bounding the x axis for better display
    :param show: string, All, pdf, cds, ppf, sample
    :param xlim: overwrites the xlim for the plots
    :return: 1
    """

    if show in ['All', 'pdf']:

        x = arange(-10, 10, 0.01)
        y = dist.pdf(x)
        plt.figure('Probability Density Function')
        plt.plot(x, y, marker='', linestyle='solid', color=color)
        if xlim:
            plt.xlim(xlim)
        plt.title(name + ' pdf')
        plt.show()

    if show in ['All', 'cdf']:
        x = arange(-10, 10, 0.01)
        y = dist.cdf(x)
        plt.figure('Cumulative Distribution Function')
        plt.plot(x, y, marker='', linestyle='solid', color=color)
        if xlim:
            plt.xlim(xlim)
        plt.title(name + ' cdf')
        plt.show()

    if show in ['All', 'ppf']:
        x = arange(0.001, 0.999, 0.001)
        y = dist.ppf(x)
        plt.figure('Quantile Function')
        plt.plot(x, y, marker='', linestyle='solid', color=color)
        plt.title(name + ' ppf')
        plt.show()

    if show in ['All', 'random_sample']:

        sample = dist.random_sample(1000)

        if bound:
            sample = sample[abs(sample) < 25]

        plt.figure('Random Sample')
        distplot(sample, bins=50, kde=False, norm_hist=True,
                 hist_kws=dict(edgecolor="white", color=color, linewidth=1.0, alpha=0.60))
        x = arange(min(sample) - 2, max(sample) + 2, 0.01)
        y = dist.pdf(x)
        plt.plot(x, y, marker='', linestyle='solid', color=color)
        if xlim:
            plt.xlim(xlim)
        plt.title(name + ' random sample')
        plt.show()

    return None


def display_parameterisations(dist=None, loc=0.0, sigma1=1.0, sigma2=1.0, sigma=1.0, gamma=0.5, show='random_sample'):
    z = dist(loc=loc, sigma1=sigma1, sigma2=sigma2)
    display_dist(dist=z, color='blue', name='standard', show=show)

    z = dist(loc=loc, sigma=sigma, gamma=gamma, kind='boe')
    display_dist(dist=z, color='coral', name='boe', show=show)

    z = dist(loc=loc, sigma=sigma, gamma=gamma, kind='inverse_scale')
    display_dist(dist=z, color='red', name='inverse_scale', show=show)

    z = dist(loc=loc, sigma=sigma, gamma=gamma, kind='percentile')
    display_dist(dist=z, color='black', name='percentile', show=show)

    return None


def display_parameterisations_shape(dist=None, loc=0.0, sigma1=1.0, sigma2=1.0,
                                    sigma=1.0, gamma=0.5, shape=2.0, show='random_sample'):
    z = dist(loc=loc, sigma1=sigma1, sigma2=sigma2, shape=shape)
    display_dist(dist=z, color='dodgerblue', name='standard', show=show)

    z = dist(loc=loc, sigma=sigma, gamma=gamma, shape=shape, kind='boe')
    display_dist(dist=z, color='coral', name='boe', show=show)

    z = dist(loc=loc, sigma=sigma, gamma=gamma, shape=shape, kind='inverse_scale')
    display_dist(dist=z, color='red', name='inverse_scale', show=show)

    z = dist(loc=loc, sigma=sigma, gamma=gamma, shape=shape, kind='percentile')
    display_dist(dist=z, color='black', name='percentile', show=show)

    return None
